/*
 * dns.c: Sending dns query and parsing response
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: jomily joseph <jomily.joseph@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <uci.h>
#include <sys/socket.h>
#include <syslog.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/tcp.h>
#include <linux/udp.h>
#include "filter.h"
#include "config.h"
#include "dns.h"

// Converts url name to a format required in dns query.
static void url_to_dns_format(unsigned char *dns, char *host)
{
	int lock = 0, i;

	strncat((char*)host, DELIM, MAX_URL_LENGTH - strlen(host));

	for (i = 0; i < (int)strlen((char*)host); i++) {
		if ('.' == host[i]) {
			*dns++ = i - lock;
			for(; lock < i; lock++) {
				*dns++ = host[lock];
			}
			lock++;
		}
	}
	*dns++ = '\0';
}

// Creates dns query and send.
// @param ip dns server ip
// @param j index of global dns cache.
// @return nothing
static void create_dns_query(char *ip, int j, bool is_ipv6)
{
	struct sockaddr_in dest;
	struct DNS_HEADER *dns = NULL;
	struct QUESTION *qinfo = NULL;
	unsigned char buf[65536], *qname;
	char url_name[MAX_URL_LENGTH];
	int s;
	struct sockaddr_in6 dest6;

	if (!is_ipv6) {
		s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		dest.sin_family = AF_INET;
		dest.sin_port = htons(DNS_PORT);
		dest.sin_addr.s_addr = inet_addr(ip);
	} else {
		s = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
		dest6.sin6_family = AF_INET6;
		dest6.sin6_port = htons(DNS_PORT);
		inet_pton(AF_INET6, ip, &dest6.sin6_addr);
	}

	// Set the DNS structure to standard queries
	dns = (struct DNS_HEADER *)&buf;
	dns->id = (unsigned short) htons(getpid());
	dns->qr = 0;
	dns->opcode = 0;
	dns->aa = 0;
	dns->tc = 0;
	dns->rd = 1;
	dns->ra = 0;
	dns->z = 0;
	dns->ad = 0;
	dns->cd = 0;
	dns->rcode = 0;
	dns->q_count = htons(1);
	dns->ans_count = 0;
	dns->auth_count = 0;
	dns->add_count = 0;

	// Point to the query portion
	qname = (unsigned char*)&buf[sizeof(struct DNS_HEADER)];

	memset(url_name, '\0', MAX_URL_LENGTH);
	strncpy(url_name, (global_dns_cache + j)->url, sizeof(url_name));
	url_to_dns_format(qname, url_name);
	qinfo = (struct QUESTION*)&buf[sizeof(struct DNS_HEADER) +
		(strlen((const char*)qname) + 1)];

	qinfo->qclass = htons(1);

	int len = sizeof(struct DNS_HEADER) +
		(strlen((const char*)qname) + 1) + sizeof(struct QUESTION);

	if (!is_ipv6) {
		qinfo->qtype = htons(1);
		if (0 > sendto(s, (char*)buf, len, 0, (struct sockaddr*)&dest, sizeof(dest))) {
			syslog(LOG_ERR, "sendto failed");
		}
	} else {
		qinfo->qtype = htons(28);
		if (0 > sendto(s, (char*)buf, len, 0, (struct sockaddr_in6*)&dest6, sizeof(dest6))) {
			syslog(LOG_ERR, "sendto failed");
		}
	}
}

#if 0
// Creates dns query and send.
// @param ipv6 dns server ipv6
// @param j index of global dns cache.
// @return nothing
static void create_ipv6_dns_query(char *ip, int j)
{
	struct sockaddr_in6 dest;
	struct DNS_HEADER *dns = NULL;
	struct QUESTION *qinfo = NULL;
	unsigned char buf[65536], *qname;
	char url_name[MAX_URL_LENGTH];
	int s;

	s = socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
	dest.sin6_family = AF_INET6;
	dest.sin6_port = htons(DNS_PORT);
	inet_pton(AF_INET6, ip, &dest.sin6_addr);

	// Set the DNS structure to standard queries
	dns = (struct DNS_HEADER *)&buf;
	dns->id = (unsigned short) htons(getpid());
	dns->qr = 0;
	dns->opcode = 0;
	dns->aa = 0;
	dns->tc = 0;
	dns->rd = 1;
	dns->ra = 0;
	dns->z = 0;
	dns->ad = 0;
	dns->cd = 0;
	dns->rcode = 0;
	dns->q_count = htons(1);
	dns->ans_count = 0;
	dns->auth_count = 0;
	dns->add_count = 0;

	// Point to the query portion
	qname = (unsigned char*)&buf[sizeof(struct DNS_HEADER)];

	memset(url_name, '\0', MAX_URL_LENGTH);
	strncpy(url_name, (global_dns_cache + j)->url, sizeof(url_name));
	url_to_dns_format(qname, url_name);
	qinfo = (struct QUESTION*)&buf[sizeof(struct DNS_HEADER) +
		(strlen((const char*)qname) + 1)];

	qinfo->qtype = htons(28);
	qinfo->qclass = htons(1);

	int len = sizeof(struct DNS_HEADER) +
		(strlen((const char*)qname) + 1) + sizeof(struct QUESTION);

	if (0 > sendto(s, (char*)buf, len, 0, (struct sockaddr_in6*)&dest, sizeof(dest))) {
		syslog(LOG_ERR, "sendto failed");
	}
}
#endif
// Sends dns query for all blacklisted url's.
void *send_dns_query(void *arg)
{
	int i, j;
	struct sockaddr_in dest;
	struct sockaddr_in6 dest6;

	// Create and send dns query to all the IPv4 dns servers.
	for (i = 0; i < dns_server_cnt; i++) {
		for (j = 0; j < dns_cache_cnt; j++) {
			if (inet_pton(AF_INET, dns_server_ip[i], &dest.sin_addr))
				create_dns_query(dns_server_ip[i], j, false);
		}
	}
	// Create and send dns query to all the IPv6 dns servers.
	for (i = 0; i < dns_server_cnt; i++) {
		for (j = 0; j < dns_cache_cnt; j++) {
			if (inet_pton(AF_INET6, dns_server_ip[i], &dest6.sin6_addr))
				create_dns_query(dns_server_ip[i], j, true);
		}
	}

	// Exit from the current thread.
	pthread_exit(NULL);
}

// Get url name from the dns format
// @param name url name
// @return nothing
static void get_url_from_dns_format(unsigned char *name)
{
	int i, j;
	unsigned int p;

	// now convert 3www6google3com0 to www.google.com.
	// The url is in dns message compression format, so we
	// just need to encode it to get the full domain name.
	for (i = 0; i < (int)strlen((const char*)name); i++) {
		p = name[i];
		for (j = 0; j < (int)p; j++) {
			name[i] = name[i + 1];
			i = i + 1;
		}
		name[i] = '.';
	}
	name[i - 1] = '\0';
}

// Read qname from the dns response packet
// @param reader pointer to read
// @param buffer contains the data
// @param count number of counts reader needs to be incremented
// @returns url name
static unsigned char *read_qname_from_dns(unsigned char *reader,
		unsigned char *buffer, int *count)
{
	unsigned char *name;
	unsigned int p = 0, jumped = 0, offset;

	*count = 1;
	name = (unsigned char*) malloc(MAX_SIZE);
	name[0] = '\0';

	// Read the names in 3www6google3com format.
	// Using the DNS name notation and message compression techniques
	// to get the qname. DNS name is encoded with a two byte subfeild,
	// which represents a pointer to a location where the name can be
	// found.
	while (0 != *reader) {
		if (192 <= *reader) {
			offset = (*reader) * MAX_SIZE + *(reader + 1) - 49152;
			reader = buffer + offset - 1;
			jumped = 1;
		} else {
			name[p++] = *reader;
		}

		reader = reader + 1;
		if (0 == jumped) {
			*count = *count + 1;
		}
	}

	name[p]='\0';
	if (1 == jumped) {
		*count = *count + 1;
	}

	get_url_from_dns_format(name);

	return name;
}

// Allocate memory for storing cname in dns cache
// @param index of global_dns_cache
// @param cnt number of cname in the response
// @return 0 on success else -1.
static int create_cname_list(int i, int cnt)
{
	int j = 0;

	(global_dns_cache + i)->cname_cnt = 0;

	(global_dns_cache + i)->cname = (char **)calloc(cnt, sizeof(char *));
	if (NULL == (global_dns_cache + i)->cname) {
		return -1;
	}

	for (j = 0; j < cnt; j++) {
		(global_dns_cache + i)->cname[j] = (char *)calloc(MAX_URL_LENGTH, sizeof(char));
		if (NULL == (global_dns_cache + i)->cname[j]) {
			if (NULL != (global_dns_cache + i)->cname) {
				free((global_dns_cache + i)->cname);
			}
			return -1;
		}
	}

	(global_dns_cache + i)->cname_cnt = cnt;

	return 0;
}

// Allocate memory for storing IPv4 address for dns cache
// @param index of global_dns_cache
// @param count number of ip address response
// @return 0 on success else -1.
static int create_ip_addr_list(int i, int count)
{
	int j = 0;

	(global_dns_cache + i)->ip_addr_cnt = 0;

	(global_dns_cache + i)->ip_addr = (char **)calloc(count, sizeof(char *));
	if (NULL == (global_dns_cache + i)->ip_addr) {
		return -1;
	}

	for (j = 0; j < count; j++) {
		(global_dns_cache + i)->ip_addr[j] = (char *)calloc(MAX_IP_LEN, sizeof(char));
		if (NULL == (global_dns_cache + i)->ip_addr[j]) {
			if (NULL != (global_dns_cache + i)->ip_addr) {
				free((global_dns_cache + i)->ip_addr);
			}
			return -1;
		}
	}

	(global_dns_cache + i)->ip_addr_cnt = count;

	return 0;
}

// Allocate memory for storing IPv6 address for dns cache
// @param index of global_dns_cache
// @param count number of ipv6 address response
// @return nothing
static int create_ipv6_addr_list(int i, int count)
{
	int j = 0;

	(global_dns_cache + i)->ipv6_addr_cnt = 0;
	(global_dns_cache + i)->ipv6_addr = (char **)calloc(count, sizeof(char *));
	if (NULL == (global_dns_cache + i)->ipv6_addr) {
		return -1;
	}

	for (j = 0; j < count; j++) {
		(global_dns_cache + i)->ipv6_addr[j] = (char *)calloc(MAX_IPV6_LEN, sizeof(char));
		if (NULL == (global_dns_cache + i)->ipv6_addr[j]) {
			if (NULL != (global_dns_cache + i)->ipv6_addr) {
				free((global_dns_cache + i)->ipv6_addr);
			}
			return -1;
		}
	}

	(global_dns_cache + i)->ipv6_addr_cnt = count;
	return 0;
}
// Function stores the IP addresses of blacklisted URL's in dns cache.
// @param dns dns header
// @param answers dns response buffer
// @param ipv4_addr_cnt number of ipv4 addresses
// @param url_cname array with cname
// @param cname_cnt number of cname per dns response.
// @returns nothing
static void store_qname_ipaddr_in_cache(struct DNS_HEADER *dns,
		struct RES_RECORD answers[MAX_DNS_ANS_RECORD], int ipv4_addr_cnt,
		char url_cname[][MAX_URL_LENGTH], int cname_cnt, int ipv6_addr_cnt)
{
	int i = 0, j = 0;
	struct sockaddr_in a;
	struct sockaddr_in6 addr_in6;
	int ipv6_addr_count = 0;

	// Check if the number of answers is not 0.
	if (0 >= ntohs(dns->ans_count)) {
		return;
	}

	// Check if the url name in dns response is same
	// as the blacklisted one, if yes then store the IP address.
	for (i = 0; i < dns_cache_cnt; i++) {
		if (NULL == strstr((char *)answers[0].name, (global_dns_cache + i)->url)) {
			continue;
		}

		// Check if memory is already present, if yes then free the memory.
		if ((0 < ipv4_addr_cnt) && (NULL != (global_dns_cache + i)->ip_addr)) {
			free((global_dns_cache + i)->ip_addr);
		}

		// Check if memory is already present, if yes then free the memory.
		if ((0 < ipv6_addr_cnt) && (NULL != (global_dns_cache + i)->ipv6_addr)) {
			free((global_dns_cache + i)->ipv6_addr);
		}

		if (NULL != (global_dns_cache + i)->cname) {
			free((global_dns_cache + i)->cname);
		}

		if ((0 < ipv4_addr_cnt) && (-1 == create_ip_addr_list(i, ipv4_addr_cnt))) {
			syslog(LOG_ERR, "Memory allocation of ipaddr list failed");
			return;
		}

		if ((0 < ipv6_addr_cnt) && (-1 == create_ipv6_addr_list(i, ipv6_addr_cnt))) {
			syslog(LOG_ERR, "Memory allocation of ipv6addr list failed");
			return;
		}

		if ((0 < cname_cnt) && (-1 == create_cname_list(i, cname_cnt))) {
			syslog(LOG_ERR, "Memory allocation of cname list failed");
			return;
		}

		for (j = 0; j < cname_cnt; j++) {
			strncpy((global_dns_cache + i)->cname[j], url_cname[j], MAX_URL_LENGTH);
		}

		int k = 0;
		for (j = 0; j < ntohs(dns->ans_count); j++) {
			if (ipv4_addr_cnt == k && (ipv6_addr_cnt == ipv6_addr_count)) {
				return;
			}
			// Check if the response has ipv4 address or not.
			if (1 == ntohs(answers[j].resource->type)) {
				long *p;
				p = (long *) answers[j].rdata;
				a.sin_addr.s_addr = (*p);
				strncpy((global_dns_cache + i)->ip_addr[k], inet_ntoa(a.sin_addr), MAX_IP_LEN);
				k++;
			}
			else if (T_AAAA == ntohs(answers[j].resource->type)) {
				unsigned char *p;
				p = (unsigned char *) answers[j].rdata;
				memcpy(&addr_in6.sin6_addr.__in6_union.__s6_addr, p, sizeof(struct in6_addr));
				inet_ntop(AF_INET6, &addr_in6.sin6_addr, (global_dns_cache + i)->ipv6_addr[ipv6_addr_count], MAX_IPV6_LEN);
				ipv6_addr_count++;
			}
		}

		return;
	}
}

// Function parses dns response.
// @param data_p payload
// @return nothing
void parse_dns_response(unsigned char *data_p)
{
	int offset, ipv4_addr = 0, i = 0, j = 0, stop = 0, cname_cnt = 0;
	struct iphdr *iph = (struct iphdr *)data_p;
	struct DNS_HEADER *dns = NULL;
	struct RES_RECORD answers[MAX_DNS_ANS_RECORD];
	unsigned char *reader, *qname;
	int ipv6_addr = 0;

	dns = (struct DNS_HEADER *) data_p;
	offset = sizeof(struct DNS_HEADER);
	qname = (unsigned char *) &data_p[offset];
	offset += (strlen((const char*)qname) + 1) + sizeof(struct QUESTION);
	reader = &data_p[offset];

	int answer_cnt = ntohs(dns->ans_count);
	char url_cname[answer_cnt][MAX_URL_LENGTH];
	memset(&url_cname, '\0', sizeof(url_cname));

	// Start reading answers in dns response
	for (i = 0; i < answer_cnt; i++) {
		answers[i].name = read_qname_from_dns(reader, data_p, &stop);
		if (NULL == answers[i].name) {
			syslog(LOG_ERR, "No qname found");
			break;
		}

		if ('\0' == answers[i].name[0]) {
			syslog(LOG_ERR, "qname empty");
			break;
		}

		reader = reader + stop;
		answers[i].resource = (struct R_DATA*)(reader);
		reader = reader + sizeof(struct R_DATA);

		// If the ip address is IPv4 then store the address
		if ((ntohs(answers[i].resource->type) == T_A) || (ntohs(answers[i].resource->type) == T_AAAA)) {
			answers[i].rdata = (unsigned char*) malloc (ntohs(answers[i].resource->data_len));
			for (j = 0; j < ntohs(answers[i].resource->data_len); j++) {
				answers[i].rdata[j] = reader[j];
			}
			answers[i].rdata[ntohs(answers[i].resource->data_len)] = '\0';
			reader = reader + ntohs(answers[i].resource->data_len);
			if (ntohs(answers[i].resource->type) == T_A)
				ipv4_addr ++;
			else
				ipv6_addr++;
		} else if (ntohs(answers[i].resource->type) == T_CNAME){
			unsigned char *cname = NULL;
			cname = read_qname_from_dns(reader, data_p, &stop);
			if (NULL != cname) {
				strncpy(url_cname[cname_cnt], (char *)cname, strlen((char *)cname));
				cname_cnt ++;
			}
			reader = reader + ntohs(answers[i].resource->data_len);
		} else {
			reader = reader + ntohs(answers[i].resource->data_len);
		}
	}

	// Store qname and ip address corresponding to it in dns cache
	if ((0 < ipv4_addr) || (0 < ipv6_addr)) {
		store_qname_ipaddr_in_cache(dns, answers, ipv4_addr, url_cname, cname_cnt, ipv6_addr);
	}
}

